"""HNN GUI visualization management tool"""

# Authors: Huzi Cheng <hzcheng15@icloud.com>

import copy
import io
from functools import partial, wraps

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from IPython.display import display
from ipywidgets import (
    Box,
    Button,
    Dropdown,
    BoundedFloatText,
    FloatText,
    HBox,
    Label,
    Layout,
    Output,
    Tab,
    VBox,
    link,
)

from hnn_core.dipole import average_dipoles, _rmse
from hnn_core.gui._logging import logger
from hnn_core.viz import plot_dipole

_fig_placeholder = "Run simulation to add figures here."

_plot_types = [
    "current dipole",
    "layer2/3 dipole",
    "layer5 dipole",
    "input histogram",
    "spikes",
    "spikes with dipoles",
    "PSD",
    "layer2/3 PSD",
    "layer5 PSD",
    "spectrogram",
    "network",
]

_no_overlay_plot_types = [
    "network",
    "spectrogram",
    "spikes",
    "spikes with dipoles",
    "input histogram",
]

_ext_data_disabled_plot_types = [
    "spikes",
    "spikes with dipoles",
    "input histogram",
    "network",
]

_spectrogram_color_maps = [
    "viridis",
    "plasma",
    "inferno",
    "magma",
    "cividis",
]

fig_templates = {
    "[Blank] 2row x 1col (1:3)": {
        "kwargs": {"gridspec_kw": {"height_ratios": [1, 3]}},
        "mosaic": "00\n11",
    },
    "[Blank] 2row x 1col (1:1)": {
        "kwargs": {"gridspec_kw": {"height_ratios": [1, 1]}},
        "mosaic": "00\n11",
    },
    "[Blank] 1row x 2col (1:1)": {
        "kwargs": {"gridspec_kw": {"height_ratios": [1, 1]}},
        "mosaic": "01\n01",
    },
    "[Blank] single figure": {
        "kwargs": {"gridspec_kw": ""},
        "mosaic": "00\n00",
    },
    "[Blank] 2row x 2col (1:1)": {
        "kwargs": {"gridspec_kw": {"height_ratios": [1, 1]}},
        "mosaic": "01\n23",
    },
}

data_templates = {
    "Drive-Dipole (2x1)": {
        "kwargs": {"gridspec_kw": {"height_ratios": [1, 3]}},
        "mosaic": "00\n11",
        "ax_plots": [("ax0", "input histogram"), ("ax1", "current dipole")],
    },
    "Dipole Layers (3x1)": {
        "kwargs": {"gridspec_kw": {"height_ratios": [1, 1, 1]}},
        "mosaic": "0\n1\n2",
        "ax_plots": [
            ("ax0", "layer2/3 dipole"),
            ("ax1", "layer5 dipole"),
            ("ax2", "current dipole"),
        ],
    },
    "Drive-Spikes (2x1)": {
        "kwargs": {"gridspec_kw": {"height_ratios": [1, 3]}},
        "mosaic": "00\n11",
        "ax_plots": [("ax0", "input histogram"), ("ax1", "spikes")],
    },
    "Dipole-Spectrogram (2x1)": {
        "kwargs": {"gridspec_kw": {"height_ratios": [1, 3]}},
        "mosaic": "00\n11",
        "ax_plots": [("ax0", "current dipole"), ("ax1", "spectrogram")],
    },
    "Dipole Layers-Spikes (1x1)": {
        "kwargs": {"gridspec_kw": ""},
        "mosaic": "00\n00",
        "ax_plots": [("ax0", "spikes with dipoles")],
    },
    "Drive-Dipole-Spectrogram (3x1)": {
        "kwargs": {"gridspec_kw": {"height_ratios": [1, 1, 2]}},
        "mosaic": "0\n1\n2",
        "ax_plots": [
            ("ax0", "input histogram"),
            ("ax1", "current dipole"),
            ("ax2", "spectrogram"),
        ],
    },
    "PSD Layers (3x1)": {
        "kwargs": {"gridspec_kw": {"height_ratios": [1, 1, 1]}},
        "mosaic": "0\n1\n2",
        "ax_plots": [
            ("ax0", "layer2/3 PSD"),
            ("ax1", "layer5 PSD"),
            ("ax2", "PSD"),
        ],
    },
}


def check_sim_plot_types(new_sim_name, plot_type_selection, target_selection, data):
    if not _is_simulation(data["simulations"][new_sim_name]):
        plot_type_selection.options = [
            pt for pt in _plot_types if pt not in _ext_data_disabled_plot_types
        ]
    else:
        plot_type_selection.options = _plot_types
    # deal with target data
    all_possible_targets = list(data["simulations"].keys())
    all_possible_targets.remove(new_sim_name)
    target_selection.options = all_possible_targets + ["None"]
    target_selection.value = "None"


def _check_template_type_is_data_dependant(template_name):
    sim_data_options = list(data_templates.keys())
    return template_name in sim_data_options


def target_comparison_change(new_target_name, simulation_selection, data):
    """Triggered when the target data is turned on or changed."""
    pass


def plot_type_coupled_change(new_plot_type, target_data_selection):
    if new_plot_type != "current dipole":
        target_data_selection.disabled = True
    else:
        target_data_selection.disabled = False


def unlink_relink(attribute):
    """
    Decorator function to unlink widgets and re-link widgets.

    Unlinks linked widgets, runs the wrapped function, and relinks the widgets
    upon completion. To be used as a decorator on class methods. The class must
    have an attribute containing an ipywidgets/traitlets link object.

    Parameters
    ----------
    attribute: (str) The class attribute containing link object of ipywidgets
               widgets

    """

    def _unlink_relink(f):
        @wraps(f)
        def wrapper(self, *args, **kwargs):
            # Unlink the widgets using the provided link object
            link_attribute: link = getattr(self, attribute)
            link_attribute.unlink()

            # Call the original function
            result = f(self, *args, **kwargs)

            # Re-link the widgets
            link_attribute.link()

            return result

        return wrapper

    return _unlink_relink


def _idx2figname(idx):
    return f"Figure {idx}"


def _figname2idx(fname):
    return int(fname.split(" ")[-1])


def _update_ax(fig, ax, single_simulation, sim_name, plot_type, plot_config):
    """Refresh plots with simulation_data.

    Parameters
    ----------
    fig : Figure
        A matplotlib.figure.Figure object.
    ax : Axes
        matplotlib.axes.Axes
    single_simulation : dict
        A single simulation
    plot_type : str
        Type of subplots
    plot_config : dict
        A dict that specifies the preprocessing and style of plots.
    """
    # Make sure that visualization does not change the original data
    dpls_copied = copy.deepcopy(single_simulation["dpls"])
    net_copied = copy.deepcopy(single_simulation["net"])
    for dpl in dpls_copied:
        if plot_config["dipole_smooth"] > 0:
            dpl.smooth(plot_config["dipole_smooth"]).scale(
                plot_config["dipole_scaling"]
            )
        else:
            dpl.scale(plot_config["dipole_scaling"])

    if net_copied is None:
        assert plot_type not in _ext_data_disabled_plot_types

    # Explicitly do this in case the
    # x and y axis are hidden after plotting some functions.
    ax.get_yaxis().set_visible(True)
    ax.get_xaxis().set_visible(True)
    if plot_type == "spikes":
        if net_copied.cell_response:
            marker_size = plot_config["marker_size"]
            hide_spike_legend = plot_config["hide_spike_legend"]
            show_legend = True if hide_spike_legend == "False" else False
            net_copied.cell_response.plot_spikes_raster(
                ax=ax,
                show=False,
                show_legend=show_legend,
                marker_size=marker_size,
            )

    if plot_type == "spikes with dipoles":
        if net_copied.cell_response:
            marker_size = plot_config["marker_size"]
            hide_spike_legend = plot_config["hide_spike_legend"]
            show_legend = True if hide_spike_legend == "False" else False
            net_copied.cell_response.plot_spikes_raster(
                ax=ax,
                show=False,
                show_legend=show_legend,
                marker_size=marker_size,
                overlay_dipoles=True,
                dpl=dpls_copied,
            )

    elif plot_type == "input histogram":
        if net_copied.cell_response:
            net_copied.external_drives.keys()

            # initialize dictionary for drives, locations
            drive_locations = dict()
            drive_colors = dict()

            for name, drive in net_copied.external_drives.items():
                # remove all increments of default 'evdist' inputs
                if "evdist" in name:
                    if "evdist" not in drive_locations.keys():
                        drive_locations["evdist"] = drive["location"]
                        drive_colors["evdist"] = "g"
                # remove all increments of default 'evprox' inputs
                elif "evprox" in name:
                    if "evprox" not in drive_locations.keys():
                        drive_locations["evprox"] = drive["location"]
                        drive_colors["evprox"] = "r"
                else:
                    drive_locations[name] = drive["location"]
                    if drive["location"] == "proximal":
                        drive_colors[name] = "r"
                    elif drive["location"] == "distal":
                        drive_colors[name] = "g"

            # all drives to plot, excluding 'evdist' and 'evprox' increments
            all_drives = list(drive_locations.keys())

            # initialize list for distal drives
            distal_drives = list()

            for name, location in drive_locations.items():
                if location == "distal":
                    distal_drives.append(name)

            net_copied.cell_response.plot_spikes_hist(
                ax=ax,
                show=False,
                spike_types=all_drives,
                invert_spike_types=distal_drives,
                color=drive_colors,
            )

    elif plot_type == "PSD":
        if len(dpls_copied) > 0:
            min_f = plot_config["min_spectral_frequency"]
            max_f = plot_config["max_spectral_frequency"]
            color = ax._get_lines.get_next_color()
            label = sim_name + " (Aggregate)"
            dpls_copied[0].plot_psd(
                fmin=min_f, fmax=max_f, color=color, label=label, ax=ax, show=False
            )

    elif plot_type == "layer2/3 PSD":
        if len(dpls_copied) > 0:
            min_f = plot_config["min_spectral_frequency"]
            max_f = plot_config["max_spectral_frequency"]
            color = ax._get_lines.get_next_color()
            label = sim_name + " (Layer 2/3)"
            dpls_copied[0].plot_psd(
                fmin=min_f,
                fmax=max_f,
                layer="L2",
                color=color,
                label=label,
                ax=ax,
                show=False,
            )

    elif plot_type == "layer5 PSD":
        if len(dpls_copied) > 0:
            min_f = plot_config["min_spectral_frequency"]
            max_f = plot_config["max_spectral_frequency"]
            color = ax._get_lines.get_next_color()
            label = sim_name + " (Layer 5)"
            dpls_copied[0].plot_psd(
                fmin=min_f,
                fmax=max_f,
                layer="L5",
                color=color,
                label=label,
                ax=ax,
                show=False,
            )

    elif plot_type == "spectrogram":
        if len(dpls_copied) > 0:
            min_f = plot_config["min_spectral_frequency"]
            max_f = plot_config["max_spectral_frequency"]
            step_f = 1.0
            if min_f > max_f:
                step_f = -1
            freqs = np.arange(min_f, max_f, step_f)
            n_cycles = freqs / 2.0

            try:
                dpls_copied[0].plot_tfr_morlet(
                    freqs,
                    n_cycles=n_cycles,
                    colormap=plot_config["spectrogram_cm"],
                    ax=ax,
                    colorbar_inside=True,
                    show=False,
                )

            except ValueError as ex:
                if str(ex) == (
                    "At least one of the wavelets is longer than "
                    "the signal. Use a longer signal or shorter "
                    "wavelets."
                ):
                    logger.error(
                        "At least one of the wavelets is "
                        "longer than the signal. Use a longer signal "
                        "or shorter wavelets. No spectrogram will be "
                        "plotted."
                    )

    elif "dipole" in plot_type:
        if len(dpls_copied) > 0:
            if len(dpls_copied) > 1:
                label = f"{sim_name}: average"
            else:
                label = sim_name

            color = ax._get_lines.get_next_color()
            if plot_type == "current dipole":
                plot_dipole(
                    dpls_copied,
                    ax=ax,
                    label=label,
                    color=color,
                    average=True,
                    show=False,
                )
            else:
                layer_namemap = {
                    "layer2/3": "L2",
                    "layer5": "L5",
                }
                plot_dipole(
                    dpls_copied,
                    ax=ax,
                    label=label,
                    color=color,
                    layer=layer_namemap[plot_type.split(" ")[0]],
                    average=True,
                    show=False,
                )
        else:
            print("No dipole data")

    elif plot_type == "network":
        if net_copied:
            with plt.ioff():
                _fig = plt.figure()
                _ax = _fig.add_subplot(111, projection="3d")
                net_copied.plot_cells(ax=_ax, show=False)
                io_buf = io.BytesIO()
                _fig.savefig(io_buf, format="raw")
                io_buf.seek(0)
                img_arr = np.reshape(
                    np.frombuffer(io_buf.getvalue(), dtype=np.uint8),
                    newshape=(int(_fig.bbox.bounds[3]), int(_fig.bbox.bounds[2]), -1),
                )
                io_buf.close()
                _ = ax.imshow(img_arr)

    # set up alignment of plots which do NOT use dipole's time as their x-axis
    custom_x_axes = [
        "network",
        "PSD",
        "layer2/3 PSD",
        "layer5 PSD",
    ]
    if plot_type not in custom_x_axes:
        margin_x = 0
        max_x = max([dpl.times[-1] for dpl in dpls_copied])
        ax.set_xlim(left=-margin_x, right=max_x + margin_x)

    return dpls_copied


def _static_rerender(widgets, fig, fig_idx):
    logger.debug("_static_re_render is called")
    figs_tabs = widgets["figs_tabs"]
    titles = figs_tabs.titles
    fig_tab_idx = titles.index(_idx2figname(fig_idx))
    fig_output = widgets["figs_tabs"].children[fig_tab_idx]
    fig_output.clear_output()
    with fig_output:
        display(fig)


def _dynamic_rerender(fig):
    fig.canvas.draw()
    fig.canvas.flush_events()


def _avg_dipole_check(dpls):
    """Check for averaged dipole, else average the trials"""
    # Check if there is an averaged dipole already
    if not dpls:
        return None

    avg_dpls = [d for d in dpls if d.nave > 1]
    if avg_dpls:
        dpl = avg_dpls[0]
    else:
        dpl = average_dipoles(dpls)
    return dpl


def _plot_on_axes(
    b,
    simulations_widget,
    widgets_plot_type,
    data_widget,
    spectrogram_colormap_selection,
    hide_spike_legend,
    marker_size,
    min_spectral_frequency,
    max_spectral_frequency,
    dipole_smooth,
    dipole_scaling,
    data_smooth,
    data_scaling,
    widgets,
    data,
    fig_idx,
    fig,
    ax,
    existing_plots,
):
    """Plotting different types of data on the given axes.

    Now this function is also responsible for comparing multiple simulations,
    or simulations vs. experimental data.

    Parameters
    ----------
    b : ipywidgets.Button
    widgets_simulation : ipywidgets.Dropdown
        A dropdown widget that contains all the simulation names.
    widgets_plot_type : ipywidgets.Dropdown
        A dropdown widget that contains all the plot types.
    target_simulations : ipywidgets.Dropdown
        The target data we want to compare with. Note that this could be 'None'
    spectrogram_colormap_selection : ipywidgets.Dropdown
        A dropdown widget that contains all the colormaps for spectrogram.
    hide_spike_legend : ipywidgets.Dropdown
        A dropdown widget that specifies whether to hide the legend for
        the spikes raster plot.
    marker_size : ipywidgets.BoundedFloatText
        A widget that specifies the marker size for the raster plot.
    dipole_smooth : ipywidgets.FloatText
        A textfield widget that specifies the smoothing window size.
    min_spectral_frequency : ipywidgets.FloatText
        A textfield that specifies the minimum frequency for spectrogram plot.
    max_spectral_frequency : ipywidgets.FloatText
        A textfield that specifies the max frequency for spectrogram plot.
    dipole_scaling : ipywidgets.FloatText
        A textfield that specifies the scaling factor for dipole object.
    widgets : dict
        A dict that contains all the widgets.
    data : dict
        A dict that contains all the simulation data. Can be accessed by names
        specified in widgets_simulation and target_simulations weidgets.
    fig_idx : int
        The index of the figure we want to plot on.
    fig : matplotlib.figure.Figure
        The figure we want to plot on.
    ax : matplotlib.axes._subplots.AxesSubplot
        The axes we want to operate.
    existing_plots : ipywidgets.VBox
        A VBox widget that contains all the existing plots.
    """
    sim_name = simulations_widget.value
    plot_type = widgets_plot_type.value
    # disable add plots for types that do not support overlay
    if plot_type in _no_overlay_plot_types:
        b.disabled = True

    # freeze plot type
    widgets_plot_type.disabled = True

    single_simulation = data["simulations"][sim_name]
    simulation_plot_config = {
        "dipole_scaling": dipole_scaling.value,
        "dipole_smooth": dipole_smooth.value,
        "min_spectral_frequency": min_spectral_frequency.value,
        "max_spectral_frequency": max_spectral_frequency.value,
        "spectrogram_cm": spectrogram_colormap_selection.value,
        "hide_spike_legend": hide_spike_legend.value,
        "marker_size": marker_size.value,
    }

    dpls_processed = _update_ax(
        fig, ax, single_simulation, sim_name, plot_type, simulation_plot_config
    )

    # If target_simulations is not None and we are plotting a dipole,
    # we need to plot the target dipole as well.
    if (
        data_widget.value in data["simulations"].keys()
        and plot_type == "current dipole"
    ):
        target_sim_name = data_widget.value
        target_sim = data["simulations"][target_sim_name]
        data_plot_config = {
            "dipole_scaling": data_scaling.value,
            "dipole_smooth": data_smooth.value,
            "min_spectral_frequency": min_spectral_frequency.value,
            "max_spectral_frequency": max_spectral_frequency.value,
            "spectrogram_cm": spectrogram_colormap_selection.value,
            "hide_spike_legend": hide_spike_legend.value,
            "marker_size": marker_size.value,
        }

        # plot the target dipole.
        target_dpl_processed = _update_ax(
            fig, ax, target_sim, target_sim_name, plot_type, data_plot_config
        )[0]  # we assume there is only one dipole.

        # calculate the RMSE between the two dipoles.
        t0 = 0.0
        tstop = dpls_processed[-1].times[-1]
        if len(dpls_processed) > 1:
            dpl = _avg_dipole_check(dpls_processed)
        else:
            dpl = dpls_processed
        rmse = _rmse(dpl, target_dpl_processed, t0, tstop)
        annotation_text = f"RMSE({sim_name}, {target_sim_name}): {rmse:.4f}"

        # find subplot's annotation
        annotation = next(
            (child for child in ax.get_children() if isinstance(child, plt.Annotation)),
            None,
        )

        # if the subplot already has an annotation, update its text.
        # Otherwise, create a new one.
        if annotation is not None:
            annotation.set_text(annotation_text)
        else:
            ax.annotate(
                annotation_text,
                xy=(0.95, 0.05),
                xycoords="axes fraction",
                horizontalalignment="right",
                verticalalignment="bottom",
                fontsize=12,
            )

        rmse_logger_text = (
            f"RMSE {rmse:.4f} ("
            f"{sim_name} smooth:{dipole_smooth.value} "
            f"scale:{dipole_scaling.value} \n"
            f"{target_sim_name} smooth:{data_smooth.value} "
            f"scale:{data_scaling.value})"
        )

        logger.info(rmse_logger_text)

    existing_plots.children = (
        *existing_plots.children,
        Label(f"{sim_name}: {plot_type}"),
    )
    if data["use_ipympl"] is False:
        _static_rerender(widgets, fig, fig_idx)
    else:
        _dynamic_rerender(fig)


def _clear_axis(
    b,
    widgets,
    data,
    fig_idx,
    fig,
    ax,
    widgets_plot_type,
    existing_plots,
    add_plot_button,
):
    ax.clear()

    # Remove "plot_spikes_hist"'s inverted second axes object, if exists, and
    # if the axis you are clearing is the spike histogram
    if ax._label == "Spike histogram":
        for axis in fig.axes:
            if axis._label == "Inverted spike histogram":
                axis.remove()

    # remove attached colorbar if exists
    if hasattr(fig, f"_cbar-ax-{id(ax)}"):
        getattr(fig, f"_cbar-ax-{id(ax)}").ax.remove()
        delattr(fig, f"_cbar-ax-{id(ax)}")

    ax.set_facecolor("w")
    ax.set_aspect("auto")
    widgets_plot_type.disabled = False
    add_plot_button.disabled = False
    existing_plots.children = ()
    if data["use_ipympl"] is False:
        _static_rerender(widgets, fig, fig_idx)
    else:
        _dynamic_rerender(fig)


def _get_ax_control(widgets, data, fig_default_params, fig_idx, fig, ax):
    analysis_style = {"description_width": "200px"}
    layout = Layout(width="98%")
    simulation_names = tuple(data["simulations"].keys())
    sim_index = 0
    default_smoothing = fig_default_params["default_smoothing"]
    default_scaling = fig_default_params["default_scaling"]
    default_min_frequency = fig_default_params["default_min_frequency"]
    default_max_frequency = fig_default_params["default_max_frequency"]
    if not simulation_names:
        simulation_names = ("None",)
    else:
        # Find the last simulation with a non-None 'net'
        sim_index = next(
            (
                idx
                for idx, sim_name in reversed(list(enumerate(simulation_names)))
                if _is_simulation(data["simulations"][sim_name])
            ),
            0,  # Default value if no such simulation is found
        )

    simulation_selection = Dropdown(
        options=simulation_names,
        value=simulation_names[sim_index],
        description="Simulation Data:",
        disabled=False,
        layout=layout,
        style=analysis_style,
    )

    plot_type_selection = Dropdown(
        options=_plot_types,
        value=_plot_types[0],
        description="Type:",
        disabled=False,
        layout=layout,
        style=analysis_style,
    )

    tagert_names = simulation_names[:-1]
    if len(simulation_names) > 1:
        tagert_names = simulation_names[1:]

    target_data_selection = Dropdown(
        options=tagert_names + ("None",),
        value="None",
        description="Data to Compare:",
        disabled=False,
        layout=layout,
        style=analysis_style,
    )

    # This will check the sim plot types dropdown available options
    # for the specific sim name in the simulation_selection dropdown options
    check_sim_plot_types(
        simulation_names[sim_index], plot_type_selection, target_data_selection, data
    )

    spectrogram_colormap_selection = Dropdown(
        description="Spectrogram Colormap:",
        options=[(cm, cm) for cm in _spectrogram_color_maps],
        value=_spectrogram_color_maps[0],
        layout=layout,
        style=analysis_style,
    )
    simulation_dipole_smooth = FloatText(
        value=default_smoothing,
        description="Dipole Smooth Window (ms):",
        disabled=False,
        layout=layout,
        style=analysis_style,
    )

    simulation_dipole_scaling = FloatText(
        value=default_scaling,
        description="Simulation Dipole Scaling:",
        disabled=False,
        layout=layout,
        style=analysis_style,
    )

    data_dipole_smooth = FloatText(
        value=0,
        description="Data Smooth Window (ms):",
        disabled=False,
        layout=layout,
        style=analysis_style,
    )

    data_dipole_scaling = FloatText(
        value=1,
        description="Data Dipole Scaling:",
        disabled=False,
        layout=layout,
        style=analysis_style,
    )

    min_spectral_frequency = BoundedFloatText(
        value=default_min_frequency,
        min=0.1,
        max=1000,
        description="Min Spectral Frequency (Hz):",
        disabled=False,
        layout=layout,
        style=analysis_style,
    )

    max_spectral_frequency = BoundedFloatText(
        value=default_max_frequency,
        min=0.1,
        max=1000,
        description="Max Spectral Frequency (Hz):",
        disabled=False,
        layout=layout,
        style=analysis_style,
    )

    hide_spike_legend = Dropdown(
        description="Hide Raster Plot Legend:",
        options=["True", "False"],
        value="False",
        layout=layout,
        style=analysis_style,
    )

    marker_size = BoundedFloatText(
        value=5,
        min=0.01,
        max=15,
        description="Raster Plot Marker Size:",
        layout=layout,
        style=analysis_style,
    )

    existing_plots = VBox([])

    plot_button = Button(description="Add plot")
    clear_button = Button(description="Clear axis")

    def _on_sim_data_change(new_sim_name):
        return check_sim_plot_types(
            new_sim_name.new, plot_type_selection, target_data_selection, data
        )

    def _on_target_comparison_change(new_target_name):
        return target_comparison_change(new_target_name, simulation_selection, data)

    def _on_plot_type_change(new_plot_type):
        return plot_type_coupled_change(new_plot_type.new, target_data_selection)

    simulation_selection.observe(_on_sim_data_change, "value")
    target_data_selection.observe(_on_target_comparison_change, "value")
    plot_type_selection.observe(_on_plot_type_change, "value")

    clear_button.on_click(
        partial(
            _clear_axis,
            widgets=widgets,
            data=data,
            fig_idx=fig_idx,
            fig=fig,
            ax=ax,
            widgets_plot_type=plot_type_selection,
            existing_plots=existing_plots,
            add_plot_button=plot_button,
        )
    )

    plot_button.on_click(
        partial(
            _plot_on_axes,
            simulations_widget=simulation_selection,
            widgets_plot_type=plot_type_selection,
            data_widget=target_data_selection,
            spectrogram_colormap_selection=spectrogram_colormap_selection,
            hide_spike_legend=hide_spike_legend,
            marker_size=marker_size,
            min_spectral_frequency=min_spectral_frequency,
            max_spectral_frequency=max_spectral_frequency,
            dipole_smooth=simulation_dipole_smooth,
            dipole_scaling=simulation_dipole_scaling,
            data_smooth=data_dipole_smooth,
            data_scaling=data_dipole_scaling,
            widgets=widgets,
            data=data,
            fig_idx=fig_idx,
            fig=fig,
            ax=ax,
            existing_plots=existing_plots,
        )
    )

    vbox = VBox(
        [
            plot_type_selection,
            simulation_selection,
            simulation_dipole_smooth,
            simulation_dipole_scaling,
            target_data_selection,
            data_dipole_smooth,
            data_dipole_scaling,
            min_spectral_frequency,
            max_spectral_frequency,
            spectrogram_colormap_selection,
            hide_spike_legend,
            marker_size,
            HBox(
                [plot_button, clear_button],
                layout=Layout(justify_content="space-between"),
            ),
            existing_plots,
        ],
        layout=Layout(width="98%"),
    )

    return vbox


def _close_figure(b, widgets, data, fig_idx):
    fig_related_widgets = [widgets["figs_tabs"], widgets["axes_config_tabs"]]
    for w_idx, tab in enumerate(fig_related_widgets):
        # Get tab object's list of children and their titles
        tab_children = list(tab.children)
        titles = list(tab.titles)
        # Get the index based on the title
        tab_idx = titles.index(_idx2figname(fig_idx))
        # Remove the child and title specified
        print(f"Del fig_idx={fig_idx}, fig_idx={fig_idx}")
        tab_children.pop(tab_idx)
        titles.pop(tab_idx)
        # Reset children and titles of the tab object
        tab.children = tab_children
        tab.titles = titles

        # If the figure tab group...
        if w_idx == 0:
            # Close figure and delete the data
            plt.close(data["figs"][fig_idx])
            data["figs"].pop(fig_idx)
            # Redisplay the remaining children
            n_tabs = len(tab.children)
            for idx in range(n_tabs):
                _fig_idx = _figname2idx(tab.get_title(idx))
                assert _fig_idx in data["figs"].keys()

                tab.children[idx].clear_output()
                with tab.children[idx]:
                    display(data["figs"][_fig_idx].canvas)

            # If all children have been deleted display the placeholder
            if n_tabs == 0:
                widgets["figs_output"].clear_output()
                with widgets["figs_output"]:
                    display(Label(_fig_placeholder))


def _add_axes_controls(widgets, data, fig_default_params, fig, axd):
    fig_idx = data["fig_idx"]["idx"]

    controls = Tab()
    children = [
        _get_ax_control(
            widgets, data, fig_default_params, fig_idx=fig_idx, fig=fig, ax=ax
        )
        for ax_key, ax in axd.items()
    ]
    controls.children = children
    for i in range(len(children)):
        controls.set_title(i, f"ax{i}")

    close_fig_button = Button(
        description=f"Close {_idx2figname(fig_idx)}",
        button_style="danger",
        icon="close",
        layout=Layout(width="98%"),
    )
    close_fig_button.on_click(
        partial(_close_figure, widgets=widgets, data=data, fig_idx=fig_idx)
    )

    n_tabs = len(widgets["axes_config_tabs"].children)
    widgets["axes_config_tabs"].children = widgets["axes_config_tabs"].children + (
        VBox([close_fig_button, controls]),
    )
    widgets["axes_config_tabs"].set_title(n_tabs, _idx2figname(fig_idx))


def _add_figure(
    b, widgets, data, fig_default_params, template_type, scale=0.95, dpi=96
):
    fig_idx = data["fig_idx"]["idx"]
    viz_output_layout = data["visualization_output"]
    fig_outputs = Output()
    n_tabs = len(widgets["figs_tabs"].children)

    if n_tabs == 0:
        widgets["figs_output"].clear_output()
        with widgets["figs_output"]:
            display(widgets["figs_tabs"])

    widgets["figs_tabs"].children = [s for s in widgets["figs_tabs"].children] + [
        fig_outputs
    ]
    widgets["figs_tabs"].set_title(n_tabs, _idx2figname(fig_idx))

    with fig_outputs:
        figsize = (
            scale * ((int(viz_output_layout.width[:-2]) - 10) / dpi),
            scale * ((int(viz_output_layout.height[:-2]) - 10) / dpi),
        )
        mosaic = template_type["mosaic"]
        kwargs = template_type["kwargs"]
        with plt.ioff():
            fig = plt.figure(figsize=figsize, dpi=dpi, layout="constrained")
            axd = fig.subplot_mosaic(mosaic, **kwargs)
            fig.canvas.header_visible = False
            fig.canvas.footer_visible = False

        if data["use_ipympl"] is False:
            plt.show()
        else:
            display(fig.canvas)

    _add_axes_controls(widgets, data, fig_default_params, fig=fig, axd=axd)

    data["figs"][fig_idx] = fig
    widgets["figs_tabs"].selected_index = n_tabs
    widgets["axes_config_tabs"].selected_index = n_tabs
    data["fig_idx"]["idx"] += 1


def _postprocess_template(template_name, fig, idx, use_ipympl=True, widgets=None):
    """Post-processes and re-renders plot templates with determined styles

    Templates are constructed on panel-by-panel basis. If adjustments need to
    be made based on information other plots in the figure, it is adjusted with
    this function. For example, L2 and L5 dipole plots should have the same
    y-axis range.
    """
    if template_name not in ["Dipole Layers (3x1)"]:
        return

    if template_name == "Dipole Layers (3x1)":
        # Make the L2, L5, and aggregate plots use the same y-range
        y_lims_list = [ax.get_ylim() for ax in fig.axes]
        y_lims = (np.min(y_lims_list), np.max(y_lims_list))
        [ax.set_ylim(y_lims) for ax in fig.axes]

    # Re-render
    if not use_ipympl:
        _static_rerender(widgets, fig, idx)
    else:
        _dynamic_rerender(fig)


class _VizManager:
    """GUI visualization panel manager class.

    Parameters
    ----------
    gui_data : dict
        A dict containing all simulation data
    viz_layout : dict
        A dict about visualization layout specs

    Attributes
    ----------
    widgets : dict
        A dict of widget references
    data : dict
        A dict of external simulation data object
    """

    def __init__(self, gui_data, viz_layout, fig_default_params):
        plt.close("all")
        self.viz_layout = viz_layout
        self.fig_default_params = fig_default_params
        self.use_ipympl = "ipympl" in matplotlib.get_backend()

        self.axes_config_output = Output()
        self.figs_output = Output()

        # widgets
        self.axes_config_tabs = Tab()
        self.figs_tabs = Tab()
        self.axes_config_tabs.selected_index = None
        self.figs_tabs.selected_index = None
        self.figs_config_tab_link = link(
            (self.axes_config_tabs, "selected_index"),
            (self.figs_tabs, "selected_index"),
        )

        template_names = list(data_templates.keys())
        template_names.extend(list(fig_templates.keys()))
        self.templates_dropdown = Dropdown(
            description="Layout template:",
            options=template_names,
            value=template_names[0],
            style={"description_width": "initial"},
            layout=Layout(width="98%"),
        )
        self.templates_dropdown.observe(self._layout_template_change, "value")

        self.make_fig_button = Button(
            description="Make figure",
            button_style="primary",
            style={"button_color": self.viz_layout["theme_color"]},
            layout=self.viz_layout["btn"],
        )
        self.make_fig_button.on_click(self.add_figure)

        self.datasets_dropdown = Dropdown(
            description="Dataset:",
            options=[],
            value=None,
            style={"description_width": "initial"},
            layout=Layout(width="98%"),
        )

        # data
        self.fig_idx = {"idx": 1}
        self.figs = {}
        self.gui_data = gui_data

    @property
    def widgets(self):
        return {
            "figs_output": self.figs_output,
            "axes_config_tabs": self.axes_config_tabs,
            "figs_tabs": self.figs_tabs,
            "templates_dropdown": self.templates_dropdown,
            "dataset_dropdown": self.datasets_dropdown,
        }

    @property
    def data(self):
        """Provides easy access to visualization-related data."""
        return {
            "use_ipympl": self.use_ipympl,
            "simulations": self.gui_data["simulation_data"],
            "fig_idx": self.fig_idx,
            "visualization_output": self.viz_layout["visualization_output"],
            "figs": self.figs,
        }

    def reset_fig_config_tabs(self, template_name=None):
        """Reset the figure config tabs with most recent simulation data."""
        simulation_names = tuple(self.data["simulations"].keys())
        for tab in self.axes_config_tabs.children:
            controls = tab.children[1]
            for ax_control in controls.children:
                # Update the options for the simulation data selection dropdown
                simulation_data_selection = ax_control.children[1]
                simulation_data_selection.options = simulation_names

                # Update the options for the data to compare dropdown
                simulation_to_compare = ax_control.children[4]
                simulation_to_compare.options = simulation_names

        # recover the default layout
        if template_name is None:
            template_name = list(fig_templates.keys())[0]
        self._simulate_switch_fig_template(template_name)

    def compose(self):
        """Compose widgets."""
        with self.axes_config_output:
            display(self.axes_config_tabs)
        with self.figs_output:
            display(Label(_fig_placeholder))

        fig_output_container = VBox(
            [self.figs_output], layout=self.viz_layout["visualization_window"]
        )

        config_panel = VBox(
            [
                Box(
                    [
                        self.templates_dropdown,
                        self.datasets_dropdown,
                        self.make_fig_button,
                    ],
                    layout=Layout(
                        display="flex",
                        flex_flow="column",
                        align_items="stretch",
                    ),
                ),
                Label("Figure config:"),
                self.axes_config_output,
            ]
        )
        return config_panel, fig_output_container

    def _layout_template_change(self, template_type):
        # check if plot set type requires loaded sim-data
        if _check_template_type_is_data_dependant(template_type.new):
            # Add only simualated data
            sim_names = [
                simulations
                for simulations, sim_name in self.data["simulations"].items()
                if sim_name["net"] is not None
            ]

            if len(sim_names) == 0:
                sim_names = [" "]

            self.datasets_dropdown.options = sim_names
            self.datasets_dropdown.value = sim_names[0]
            # show list of simulated to gui dropdown
            self.datasets_dropdown.layout.visibility = "visible"
        else:
            # hide sim-data dropdown
            self.datasets_dropdown.layout.visibility = "hidden"

    @unlink_relink(attribute="figs_config_tab_link")
    def add_figure(self, b=None):
        """Add a figure and corresponding config tabs to the dashboard."""
        if len(self.data["simulations"]) == 0:
            logger.error("No data has been loaded")
            return

        template_name = self.widgets["templates_dropdown"].value
        is_data_template = _check_template_type_is_data_dependant(template_name)
        if is_data_template:
            sim_name = self.widgets["dataset_dropdown"].value
            if sim_name not in self.data["simulations"]:
                logger.error("No simulation data has been loaded")
                return

        # Use data_templates dictionary if it's a data dependent layout
        template_type = (
            data_templates[template_name]
            if is_data_template
            else fig_templates[template_name]
        )

        # Add empty figure according to template arguments
        _add_figure(
            None,
            self.widgets,
            self.data,
            self.fig_default_params,
            template_type,
            scale=0.97,
            dpi=self.viz_layout["dpi"],
        )

        # Plot data if it is a data-dependent template
        if is_data_template:
            fig_name = _idx2figname(self.data["fig_idx"]["idx"] - 1)
            # get figs per axis
            ax_plots = data_templates[template_name]["ax_plots"]
            for ax_name, plot_type in ax_plots:
                # paint fig in axis
                self._simulate_edit_figure(
                    fig_name, ax_name, sim_name, plot_type, {}, "plot"
                )
            # template post-processing
            fig_key = self.data["fig_idx"]["idx"] - 1
            _postprocess_template(
                template_name,
                fig=self.figs[fig_key],
                idx=fig_key,
                use_ipympl=self.use_ipympl,
                widgets=self.widgets,
            )

            logger.info(
                f"Figure {template_name} for simulation {sim_name} has been created"
            )

    def _simulate_add_fig(self):
        self.make_fig_button.click()

    def _simulate_switch_fig_template(self, template_name):
        assert template_name in fig_templates.keys() or data_templates.keys(), (
            "No such template"
        )
        self.templates_dropdown.value = template_name

    def _simulate_delete_figure(self, fig_name):
        tab = self.axes_config_tabs
        titles = tab.titles
        assert fig_name in titles
        tab_idx = titles.index(fig_name)

        self.axes_config_tabs.selected_index = tab_idx
        close_button = self.axes_config_tabs.children[tab_idx].children[0]
        close_button.click()

    def _simulate_edit_figure(
        self,
        fig_name,
        ax_name,
        simulation_name,
        plot_type,
        preprocessing_config,
        operation,
    ):
        """Manipulate a certain figure.

        Parameters
        ----------
            fig_name : str
                The figure name shown in the GUI, e.g., 'Figure 1'.
            ax_name : str
                Axis name shown in the left side of GUI, like, 'ax0'.
            simulation_name : str
                The name of simulation you want to visualize
            plot_type : str
                Type of visualization.
            preprocessing_config : dict
                A dict of visualization preprocessing parameters. Allowed keys:
                `dipole_smooth`, `dipole_scaling`,
                `data_to_compare`, `data_smooth`, `data_scaling`
                `min_spectral_frequency`, `max_spectral_frequency`,
                `spectrogram_colormap_selection`, `hide_spike_legend,
                `marker_size`.
                config could be empty: `{}`.
            operation : str
                `"plot"` if you want to plot and `"clear"` if you want to
                remove previously plotted visualizations.
        """
        assert simulation_name in self.data["simulations"].keys()
        assert plot_type in _plot_types
        assert operation in ("plot", "clear")

        # Select the figure tab
        tab = self.axes_config_tabs
        titles = tab.titles
        assert fig_name in titles, "No such figure"
        tab_idx = titles.index(fig_name)
        self.axes_config_tabs.selected_index = tab_idx

        # Select the figure panel/ax tab
        ax_control_tabs = self.axes_config_tabs.children[tab_idx].children[1]
        ax_titles = ax_control_tabs.titles
        assert ax_name in ax_titles, "No such axis"
        ax_idx = ax_titles.index(ax_name)
        ax_control_tabs.selected_index = ax_idx

        # Select the simulation
        simulation_selector = ax_control_tabs.children[ax_idx].children[1]
        simulation_selector.value = simulation_name

        # Select the plot type
        plot_type_selector = ax_control_tabs.children[ax_idx].children[0]
        plot_type_selector.value = plot_type

        # Set the plot configurations
        config_name_idx = {
            "dipole_smooth": 2,
            "dipole_scaling": 3,
            "data_to_compare": 4,
            "data_smooth": 5,
            "data_scaling": 6,
            "min_spectral_frequency": 7,
            "max_spectral_frequency": 8,
            "spectrogram_colormap_selection": 9,
            "hide_spike_legend": 10,
            "marker_size": 11,
        }
        for conf_key, conf_val in preprocessing_config.items():
            assert conf_key in config_name_idx.keys()
            idx = config_name_idx[conf_key]
            conf_widget = ax_control_tabs.children[ax_idx].children[idx]
            conf_widget.value = conf_val

        buttons = ax_control_tabs.children[ax_idx].children[-2]
        if operation == "plot":
            buttons.children[0].click()
        elif operation == "clear":
            buttons.children[1].click()


def _is_simulation(data):
    """Determines if saved data is a simulation."""
    return data["net"] is not None



"""
============================================
01. Simulate Event Related Potentials (ERPs)
============================================

This example demonstrates how to simulate a threshold level tactile
evoked response, as detailed in the `HNN GUI ERP tutorial
<https://jonescompneurolab.github.io/hnn-tutorials/erp/erp>`_,
using HNN-core. We recommend you first review the GUI tutorial.

The workflow below recreates an example of the threshold level tactile
evoked response, as observed in Jones et al. J. Neuroscience 2007 [1]_
(e.g. Figure 7 in the GUI tutorial), albeit without a direct comparison
to the recorded data.
"""

# Authors: Mainak Jas <mmjas@mgh.harvard.edu>
#          Sam Neymotin <samnemo@gmail.com>
#          Blake Caldwell <blake_caldwell@brown.edu>
#          Christopher Bailey <cjb@cfin.au.dk>

# sphinx_gallery_thumbnail_number = 3

import os.path as op
import tempfile

import matplotlib.pyplot as plt

###############################################################################
# Let us import hnn_core

import hnn_core
from hnn_core import simulate_dipole, jones_2009_model
from hnn_core.viz import plot_dipole

###############################################################################
# Let us first create our default network and visualize the cells
# inside it.
net = jones_2009_model()
net.plot_cells()
net.cell_types['L5_pyramidal']['cell_object'].plot_morphology()

###############################################################################
# The network of cells is now defined, to which we add external drives as
# required. Weights are prescribed separately for AMPA and NMDA receptors
# (receptors that are not used can be omitted or set to zero). The possible
# drive types include the following (click on the links for documentation):
#
# - :meth:`hnn_core.Network.add_evoked_drive`
# - :meth:`hnn_core.Network.add_poisson_drive`
# - :meth:`hnn_core.Network.add_bursty_drive`

###############################################################################
# First, we add a distal evoked drive
weights_ampa_d1 = {'L2_basket': 0.006562, 'L2_pyramidal': .000007,
                   'L5_pyramidal': 0.142300}
weights_nmda_d1 = {'L2_basket': 0.019482, 'L2_pyramidal': 0.004317,
                   'L5_pyramidal': 0.080074}
synaptic_delays_d1 = {'L2_basket': 0.1, 'L2_pyramidal': 0.1,
                      'L5_pyramidal': 0.1}
net.add_evoked_drive(
    'evdist1', mu=63.53, sigma=3.85, numspikes=1, weights_ampa=weights_ampa_d1,
    weights_nmda=weights_nmda_d1, location='distal',
    synaptic_delays=synaptic_delays_d1, event_seed=274)

###############################################################################
# Then, we add two proximal drives
weights_ampa_p1 = {'L2_basket': 0.08831, 'L2_pyramidal': 0.01525,
                   'L5_basket': 0.19934, 'L5_pyramidal': 0.00865}
synaptic_delays_prox = {'L2_basket': 0.1, 'L2_pyramidal': 0.1,
                        'L5_basket': 1., 'L5_pyramidal': 1.}
# all NMDA weights are zero; pass None explicitly
net.add_evoked_drive(
    'evprox1', mu=26.61, sigma=2.47, numspikes=1, weights_ampa=weights_ampa_p1,
    weights_nmda=None, location='proximal',
    synaptic_delays=synaptic_delays_prox, event_seed=544)

# Second proximal evoked drive. NB: only AMPA weights differ from first
weights_ampa_p2 = {'L2_basket': 0.000003, 'L2_pyramidal': 1.438840,
                   'L5_basket': 0.008958, 'L5_pyramidal': 0.684013}
# all NMDA weights are zero; omit weights_nmda (defaults to None)
net.add_evoked_drive(
    'evprox2', mu=137.12, sigma=8.33, numspikes=1,
    weights_ampa=weights_ampa_p2, location='proximal',
    synaptic_delays=synaptic_delays_prox, event_seed=814)

###############################################################################
# Now let's simulate the dipole, running 2 trials with the
# :class:`~hnn_core.parallel_backends.Joblib` backend.
# To run them in parallel we could set ``n_jobs`` to equal the number of
# trials. The ``Joblib`` backend allows running the simulations in parallel
# across trials.
from hnn_core import JoblibBackend

with JoblibBackend(n_jobs=2):
    dpls = simulate_dipole(net, tstop=170., n_trials=2)

###############################################################################
# Rather than reading smoothing and scaling parameters from file, we recommend
# explicit use of the :meth:`~hnn_core.dipole.Dipole.smooth` and
# :meth:`~hnn_core.dipole.Dipole.scale` methods instead. Note that both methods
# operate in-place, i.e., the objects are modified.
window_len, scaling_factor = 30, 3000
for dpl in dpls:
    dpl.smooth(window_len).scale(scaling_factor)

###############################################################################
# Plot the amplitudes of the simulated aggregate dipole moments over time
import matplotlib.pyplot as plt
fig, axes = plt.subplots(2, 1, sharex=True, figsize=(6, 6),
                         constrained_layout=True)
plot_dipole(dpls, ax=axes[0], layer='agg', show=False)
net.cell_response.plot_spikes_hist(ax=axes[1],
                                   spike_types=['evprox', 'evdist'])

###############################################################################
# If you want to analyze how the different cortical layers contribute to
# different net waveform features, then instead of passing ``'agg'`` to
# ``layer``, you can provide a list of layers to be visualized and optionally
# a list of axes to ``ax`` to visualize the dipole moments separately.
plot_dipole(dpls, average=False, layer=['L2', 'L5', 'agg'], show=False)


###############################################################################
# Now, let us try to make the exogenous driving inputs to the cells
# synchronous and see what happens. This is achieved by setting
# ``n_drive_cells=1`` and ``cell_specific=False`` when adding each drive.

net_sync = jones_2009_model()

n_drive_cells=1
cell_specific=False

net_sync.add_evoked_drive(
    'evdist1', mu=63.53, sigma=3.85, numspikes=1, weights_ampa=weights_ampa_d1,
    weights_nmda=weights_nmda_d1, location='distal', n_drive_cells=n_drive_cells,
    cell_specific=cell_specific, synaptic_delays=synaptic_delays_d1, event_seed=274)

net_sync.add_evoked_drive(
    'evprox1', mu=26.61, sigma=2.47, numspikes=1, weights_ampa=weights_ampa_p1,
    weights_nmda=None, location='proximal', n_drive_cells=n_drive_cells,
    cell_specific=cell_specific, synaptic_delays=synaptic_delays_prox, event_seed=544)

net_sync.add_evoked_drive(
    'evprox2', mu=137.12, sigma=8.33, numspikes=1,
    weights_ampa=weights_ampa_p2, location='proximal', n_drive_cells=n_drive_cells,
    cell_specific=cell_specific, synaptic_delays=synaptic_delays_prox, event_seed=814)

###############################################################################
# You may interrogate current values defining the spike event time dynamics by
print(net_sync.external_drives['evdist1']['dynamics'])

###############################################################################
# Finally, let's simulate this network. Rather than modifying the dipole
# object, this time we make a copy of it before smoothing and scaling.
dpls_sync = simulate_dipole(net_sync, tstop=170., n_trials=1)

trial_idx = 0
dpls_sync[trial_idx].copy().smooth(window_len).scale(scaling_factor).plot()
net_sync.cell_response.plot_spikes_hist()

###############################################################################
# .. warning::
#    - Always look at dipoles in conjunction with raster plots and spike
#      histogram to avoid misinterpretation.
#    - Run multiple trials of your simulation to get an average of different
#      drives seeds before drawing conclusions.

###############################################################################
# References
# ----------
# .. [1] Jones, Stephanie R., et al. "Neural correlates of tactile detection:
#        a combined magnetoencephalography and biophysically based computational
#        modeling study." Journal of Neuroscience 27.40 (2007): 10751-10764.



"""Objective functions for parameter optimization."""

# Authors: Carolina Fernandez <cxf418@miami.edu>
#          Nick Tolley <nicholas_tolley@brown.edu>
#          Ryan Thorpe <ryan_thorpe@brown.edu>
#          Mainak Jas <mjas@mgh.harvard.edu>

from hnn_core import simulate_dipole
from ..dipole import _rmse


def _rmse_evoked(
    initial_net,
    initial_params,
    set_params,
    predicted_params,
    update_params,
    obj_values,
    tstop,
    obj_fun_kwargs,
):
    """The objective function for evoked responses.

    Parameters
    ----------
    initial_net : instance of Network
        The network object.
    initial_params : dict
        Keys are parameter names, values are initial parameters.
    set_params : func
        User-defined function that sets network drives and parameters.
    predicted_params : list
        Parameters selected by the optimizer.
    update_params : func
        Function to update params.
    tstop : float
        The simulated dipole's duration.
    target : instance of Dipole
        A dipole object with experimental data.

    Returns
    -------
    obj : float
        Normalized RMSE between recorded and simulated dipole.
    """

    params = update_params(initial_params, predicted_params)

    # simulate dpl with predicted params
    new_net = initial_net.copy()
    set_params(new_net, params)
    dpl = simulate_dipole(new_net, tstop=tstop, n_trials=1)[0]

    # smooth & scale
    if "scale_factor" in obj_fun_kwargs:
        dpl.scale(obj_fun_kwargs["scale_factor"])
    if "smooth_window_len" in obj_fun_kwargs:
        dpl.smooth(obj_fun_kwargs["smooth_window_len"])

    obj = _rmse(dpl, obj_fun_kwargs["target"], tstop=tstop)

    obj_values.append(obj)

    return obj


def _maximize_psd(
    initial_net,
    initial_params,
    set_params,
    predicted_params,
    update_params,
    obj_values,
    tstop,
    obj_fun_kwargs,
):
    """The objective function for PSDs.

    Parameters
    ----------
    initial_net : instance of Network
        The network object.
    initial_params : dict
        Keys are parameter names, values are initial parameters.
    set_params : func
        User-defined function that sets network drives and parameters.
    predicted_params : list
        Parameters selected by the optimizer.
    update_params : func
        Function to update params.
    tstop : float
        The simulated dipole's duration.
    f_bands : list of tuples
        Lower and higher limit for each frequency band.
    relative_bandpower : tuple
        Weight for each frequency band.

    Returns
    -------
    obj : float
        Sum of the weighted frequency band PSDs relative to total signal PSD.

    Notes
    -----
    The objective function minimizes the sum of the weighted (user-defined)
    frequency band PSDs (user-defined) relative to the total PSD of the signal.
    The objective function can be represented as -Σc[ΣPSD(i)/ΣPSD(j)] where c
    is the weight for each frequency band, PSD(i) is the PSD for each frequency
    band, and PSD(j) is the total PSD of the signal.
    """

    import numpy as np

    from scipy.signal import periodogram

    params = update_params(initial_params, predicted_params)

    # simulate dpl with predicted params
    new_net = initial_net.copy()
    set_params(new_net, params)
    dpl = simulate_dipole(new_net, tstop=tstop, n_trials=1)[0]

    # smooth & scale
    if "scale_factor" in obj_fun_kwargs:
        dpl.scale(obj_fun_kwargs["scale_factor"])
    if "smooth_window_len" in obj_fun_kwargs:
        dpl.smooth(obj_fun_kwargs["smooth_window_len"])

    # resample?

    # get psd of simulated dpl
    freqs_simulated, psd_simulated = periodogram(
        dpl.data["agg"], dpl.sfreq, window="hamming"
    )

    # for each f band
    f_bands_psds = list()
    for idx, f_band in enumerate(obj_fun_kwargs["f_bands"]):
        f_band_idx = np.where(
            np.logical_and(freqs_simulated >= f_band[0], freqs_simulated <= f_band[1])
        )[0]
        f_bands_psds.append(
            -obj_fun_kwargs["relative_bandpower"][idx] * sum(psd_simulated[f_band_idx])
        )

    # grand sum
    obj = sum(f_bands_psds) / sum(psd_simulated)

    obj_values.append(obj)

    return obj



from .dipole import (
    simulate_dipole,
    read_dipole,
    average_dipoles,
    Dipole,
    _read_dipole_txt,
)
from .params import Params, read_params, convert_to_json
from .network import Network, pick_connection
from .network_models import jones_2009_model, law_2021_model, calcium_model
from .cell import Cell
from .cell_response import CellResponse, read_spikes
from .cells_default import pyramidal, basket
from .parallel_backends import MPIBackend, JoblibBackend
from .hnn_io import (
    dict_to_network,
    network_to_dict,
    read_network_configuration,
    write_network_configuration,
)

__version__ = "0.4.3"



"""
=======================
01. Plot firing pattern
=======================

This example demonstrates how to inspect the firing
pattern of cells in the HNN model.
"""

# Authors: Mainak Jas <mjas@harvard.mgh.edu>
#          Nick Tolley <nick nicholas_tolley@brown.edu>

import os.path as op
import tempfile

###############################################################################
# Let us import ``hnn_core``.

import hnn_core
from hnn_core import read_spikes, jones_2009_model, simulate_dipole

###############################################################################
# Now let's build the network. We have used the same weights as in the
# :ref:`evoked example <sphx_glr_auto_examples_plot_simulate_evoked.py>`.
import matplotlib.pyplot as plt

net = jones_2009_model()

###############################################################################
# ``net`` does not have any driving inputs and only defines the local network
# connectivity. Let us go ahead and first add a distal evoked drive.
# We need to define the AMPA and NMDA weights for the connections. An
# "evoked drive" defines inputs that are normally distributed with a certain
# mean and standard deviation.

weights_ampa_d1 = {'L2_basket': 0.006562, 'L2_pyramidal': 7e-6,
                   'L5_pyramidal': 0.142300}
weights_nmda_d1 = {'L2_basket': 0.019482, 'L2_pyramidal': 0.004317,
                   'L5_pyramidal': 0.080074}
synaptic_delays_d1 = {'L2_basket': 0.1, 'L2_pyramidal': 0.1,
                      'L5_pyramidal': 0.1}
net.add_evoked_drive(
    'evdist1', mu=63.53, sigma=3.85, numspikes=1, weights_ampa=weights_ampa_d1,
    weights_nmda=weights_nmda_d1, location='distal',
    synaptic_delays=synaptic_delays_d1, event_seed=274)

###############################################################################
# The reason it is called an "evoked drive" is it can be used to simulate
# waveforms resembling evoked responses. Here, we show how to do it with two
# proximal drives which drive current up the dendrite and one distal drive
# which drives current down the dendrite producing the negative deflection.
weights_ampa_p1 = {'L2_basket': 0.08831, 'L2_pyramidal': 0.01525,
                   'L5_basket': 0.19934, 'L5_pyramidal': 0.00865}
synaptic_delays_prox = {'L2_basket': 0.1, 'L2_pyramidal': 0.1,
                        'L5_basket': 1., 'L5_pyramidal': 1.}

# all NMDA weights are zero; pass None explicitly
net.add_evoked_drive(
    'evprox1', mu=26.61, sigma=2.47, numspikes=1, weights_ampa=weights_ampa_p1,
    weights_nmda=None, location='proximal',
    synaptic_delays=synaptic_delays_prox, event_seed=544)

###############################################################################
# Now we add the second proximal evoked drive and simulate the network
# dynamics with somatic voltage recordings enabled. Note: only AMPA weights
# differ from first.
weights_ampa_p2 = {'L2_basket': 0.000003, 'L2_pyramidal': 1.438840,
                   'L5_basket': 0.008958, 'L5_pyramidal': 0.684013}
# all NMDA weights are zero; omit weights_nmda (defaults to None)
net.add_evoked_drive(
    'evprox2', mu=137.12, sigma=8.33, numspikes=1,
    weights_ampa=weights_ampa_p2, location='proximal',
    synaptic_delays=synaptic_delays_prox, event_seed=814)

dpls = simulate_dipole(net, tstop=170., record_vsec='soma')

###############################################################################
# Here, we explain more details about the data structures and how they can
# be used to better interpret the data. The cell IDs (gids) uniquely define
# neurons in the network and are stored in the :class:`~hnn_core.Network`
# object as a dictionary
gid_ranges = net.gid_ranges
print(net.gid_ranges)

###############################################################################
# Simulated voltage in the soma and other cell sections are stored in
# :class:`~hnn_core.CellResponse` as a dictionary. The CellResponse object
# stores data produced by individual cells including spikes,
# voltages and currents.
trial_idx = 0
vsec = net.cell_response.vsec[trial_idx]
print(vsec.keys())

###############################################################################
# We can plot the firing pattern of individual cells by indexing with the gid
gid = 170
plt.figure(figsize=(4, 4), constrained_layout=True)
plt.plot(net.cell_response.times, vsec[gid]['soma'])
plt.title('%s (gid=%d)' % (net.gid_to_type(gid), gid))
plt.xlabel('Time (ms)')
plt.ylabel('Voltage (mV)')
plt.show()

###############################################################################
# Also, we can plot the spikes in the network and write them to text files.
# Note that we can use formatting syntax to specify the filename pattern
# with which each trial will be written ('spk_1.txt', 'spk_2.txt, ...). To
# read spikes back in, we can use wildcard expressions.
net.cell_response.plot_spikes_raster()
with tempfile.TemporaryDirectory() as tmp_dir_name:
    net.cell_response.write(op.join(tmp_dir_name, 'spk_%d.txt'))
    cell_response = read_spikes(op.join(tmp_dir_name, 'spk_*.txt'))
cell_response.plot_spikes_raster()

###############################################################################
# We can additionally calculate the mean spike rates for each cell class by
# specifying a time window with ``tstart`` and ``tstop``.
all_rates = cell_response.mean_rates(tstart=0, tstop=170,
                                     gid_ranges=net.gid_ranges,
                                     mean_type='all')
trial_rates = cell_response.mean_rates(tstart=0, tstop=170,
                                       gid_ranges=net.gid_ranges,
                                       mean_type='trial')
print('Mean spike rates across trials:')
print(all_rates)
print('Mean spike rates for individual trials:')
print(trial_rates)

###############################################################################
# Finally, we can plot the soma voltage along with the spiking activity with
# raster plots and histograms for the pyramidal cells.

fig, axes = plt.subplots(3, 1, figsize=(5, 7), sharex=True)

for idx in range(10):  # only 10 cells per cell-type
    gid = gid_ranges['L2_pyramidal'][idx]
    axes[0].plot(net.cell_response.times, vsec[gid]['soma'], color='g')
    gid = gid_ranges['L5_pyramidal'][idx]
    axes[0].plot(net.cell_response.times, vsec[gid]['soma'], color='r')
net.cell_response.plot_spikes_raster(ax=axes[1])
net.cell_response.plot_spikes_hist(ax=axes[2],
                                   spike_types=['L5_pyramidal',
                                                'L2_pyramidal'])



"""
================================
03. Modifying local connectivity
================================

This example demonstrates how to modify the network connectivity.
"""

# Author: Nick Tolley <nicholas_tolley@brown.edu>

# sphinx_gallery_thumbnail_number = 2

import os.path as op

###############################################################################
# Let us import ``hnn_core``.

import hnn_core
from hnn_core import jones_2009_model, simulate_dipole

###############################################################################
# To explore how to modify network connectivity, we will start with simulating
# the evoked response from the
# :ref:`evoked example <sphx_glr_auto_examples_plot_simulate_evoked.py>`, and
# explore how it changes with new connections. We first instantiate the
# network. (Note: Setting ``add_drives_from_params=True`` loads a set of
# predefined drives without the drives API shown previously).
net_erp = jones_2009_model(add_drives_from_params=True)

###############################################################################
# Instantiating the network comes with a predefined set of connections that
# reflect the canonical neocortical microcircuit. ``net.connectivity``
# is a list of dictionaries which detail every cell-cell, and drive-cell
# connection. The weights of these connections can be visualized with
# :func:`~hnn_core.viz.plot_connectivity_weights` as well as
# :func:`~hnn_core.viz.plot_cell_connectivity`. We can search for specific
# connections using ``pick_connection`` which returns the indices
# of ``net.connectivity`` that match the provided parameters.
from hnn_core.viz import plot_connectivity_matrix, plot_cell_connectivity, plot_drive_strength
from hnn_core.network import pick_connection

print(len(net_erp.connectivity))

conn_indices = pick_connection(
    net=net_erp, src_gids='L5_basket', target_gids='L5_pyramidal',
    loc='soma', receptor='gabaa')
conn_idx = conn_indices[0]
print(net_erp.connectivity[conn_idx])
plot_connectivity_matrix(net_erp, conn_idx)

# Note here that `'src_gids'` is a `set` object
# The `.pop()` method can be used to sample a random element
src_gid = net_erp.connectivity[conn_idx]['src_gids'].copy().pop()
fig = plot_cell_connectivity(net_erp, conn_idx, src_gid)

###############################################################################
# Data recorded during simulations are stored under
# :class:`~hnn_core.Cell_Response`. Spiking activity can be visualized after
# a simulation is using :meth:`~hnn_core.Cell_Response.plot_spikes_raster`
dpl_erp = simulate_dipole(net_erp, tstop=170., n_trials=1)
net_erp.cell_response.plot_spikes_raster()

###############################################################################
# We can also define our own connections to test the effect of different
# connectivity patterns. To start, ``net.clear_connectivity()`` can be used
# to clear all cell-to-cell connections. By default, previously defined drives
# to the network are retained, but can be removed with ``net.clear_drives()``.
# ``net.add_connection`` is then used to create a custom network. Let us first
# create an all-to-all connectivity pattern between the L5 pyramidal cells,
# and L2 basket cells. :meth:`hnn_core.Network.add_connection` allows
# connections to be specified with either cell names, or the cell IDs (gids)
# directly.
def get_network(probability=1.0):
    net = jones_2009_model(add_drives_from_params=True)
    net.clear_connectivity()

    # Pyramidal cell connections
    location, receptor = 'distal', 'ampa'
    weight, delay, lamtha = 1.0, 1.0, 70
    src = 'L5_pyramidal'
    conn_seed = 3
    for target in ['L5_pyramidal', 'L2_basket']:
        net.add_connection(src, target, location, receptor,
                           delay, weight, lamtha, probability=probability,
                           conn_seed=conn_seed)

    # Basket cell connections
    location, receptor = 'soma', 'gabaa'
    weight, delay, lamtha = 1.0, 1.0, 70
    src = 'L2_basket'
    for target in ['L5_pyramidal', 'L2_basket']:
        net.add_connection(src, target, location, receptor,
                           delay, weight, lamtha, probability=probability,
                           conn_seed=conn_seed)
    return net


net_all = get_network()
dpl_all = simulate_dipole(net_all, tstop=170., n_trials=1)

###############################################################################
# We can additionally use the ``probability`` argument to create a sparse
# connectivity pattern instead of all-to-all. Let's try creating the same
# network with a 10% chance of cells connecting to each other.
net_sparse = get_network(probability=0.1)
dpl_sparse = simulate_dipole(net_sparse, tstop=170., n_trials=1)

###############################################################################
# With the previous connection pattern there appears to be synchronous rhythmic
# firing of the L5 pyramidal cells with a period of 10 ms. The synchronous
# activity is visible as vertical lines where several cells fire simultaneously
# Using the sparse connectivity pattern produced a lot more spiking in
# the L5 pyramidal cells.
net_all.cell_response.plot_spikes_raster()
net_sparse.cell_response.plot_spikes_raster()

###############################################################################
# We can plot the sparse connectivity pattern between cell populations.
conn_indices = pick_connection(
    net=net_sparse, src_gids='L2_basket', target_gids='L2_basket',
    loc='soma', receptor='gabaa')

conn_idx = conn_indices[0]
plot_connectivity_matrix(net_sparse, conn_idx)

###############################################################################
# Note that the sparsity is in addition to the weight decay with distance
# from the source cell.
src_gid = net_sparse.connectivity[conn_idx]['src_gids'].copy().pop()
plot_cell_connectivity(net_sparse, conn_idx, src_gid=src_gid)

###############################################################################
# We can plot the strengths of each external drive across each cell types.

# This can be done in a relative way, enabling you to compare the proportion
# coming from each drive:
plot_drive_strength(net_erp)

# Alternatively, you can compare the total amount of each drive in conductance:
plot_drive_strength(net_erp, normalize=False)

###############################################################################
# In the sparse network, there still appears to be some rhythmicity
# where the cells are firing synchronously with a smaller period of 4-5 ms.
# As a final step, we can see how this change in spiking activity impacts
# the aggregate current dipole.
import matplotlib.pyplot as plt
from hnn_core.viz import plot_dipole
fig, axes = plt.subplots(2, 1, sharex=True, figsize=(6, 6),
                         constrained_layout=True)

window_len = 30  # ms
scaling_factor = 3000
dpls = [dpl_erp[0].smooth(window_len).scale(scaling_factor),
        dpl_all[0].smooth(window_len).scale(scaling_factor),
        dpl_sparse[0].smooth(window_len).scale(scaling_factor)]

plot_dipole(dpls, ax=axes[0], layer='agg', show=False)
axes[0].legend(['Default', 'Custom All', 'Custom Sparse'])
net_erp.cell_response.plot_spikes_hist(
    ax=axes[1], spike_types=['evprox', 'evdist'])
